/*
  */

#include "ocr.h"

#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>
#include <X11/keysym.h>

CanonImage *page_images[MAX_PAGES];
int npages;

static Window id;
static Display *disp;
static struct timeval tv_startup;

static KeyCode keycode(KeySym sym) {
  return XKeysymToKeycode(disp,sym);
}

static void check_pclose(FILE *f, char *cmd) {
  int r;
  eassert(!ferror(f));
  r= fgetc(f);  eassert(r==EOF);  eassert(feof(f));
  r= pclose(f);  eassert(r>=0);  eassert(WIFEXITED(r) && !WEXITSTATUS(r));
  free(cmd);
}

#if 0
static CanonImage *screenshot_now(void) {
  char *cmd;
  CanonImage *ci;
  int r;
  
  r= asprintf(&cmd, "xwd -silent -id 0x%lx | xwdtopnm", (unsigned long)id);
  eassert(r>=0);
  FILE *f= popen(cmd,"r");  eassert(f);
  ci= file_read_image(f);
  check_pclose(f, cmd);
  return ci;
}
#endif

static void screenshot_startup(void) {
  int r;
  disp= XOpenDisplay(0);  eassert(disp);
  r= gettimeofday(&tv_startup,0);  eassert(!r);
}

#if 0
static CanonImage *single_page(void) {
  int r;
  r= XRaiseWindow(disp, id);  eassert(r);
  r= XSync(disp, False);  eassert(r);
  return screenshot_now();
}
#endif

static void raise_and_set_focus(void) {
  int r;
  XWindowAttributes attr;
  int xpos,ypos, evbase,errbase,majver,minver;
  unsigned width,height,bd,depth;
  Window dummy;
  
  fprintf(stderr,"PAGING raise_and_set_focus\n");

  r= XTestQueryExtension(disp, &evbase,&errbase,&majver,&minver);
  eassert(r==True);

  r= XRaiseWindow(disp, id);  eassert(r);

  r= XGetWindowAttributes(disp, id, &attr);  eassert(r);
  r= XGetGeometry(disp,id, &attr.root,&xpos,&ypos,&width,&height, &bd,&depth);
  eassert(r);

  r= XTranslateCoordinates(disp, id,attr.root, 160,160, &xpos,&ypos,
			   &dummy);
  eassert(r);

  int screen= XScreenNumberOfScreen(attr.screen);
  XTestFakeMotionEvent(disp,screen, xpos, ypos, 0);

  XTestFakeButtonEvent(disp,1,1, 50);
  XTestFakeButtonEvent(disp,1,0, 50);
  r= XSync(disp, False);  eassert(r);

  fprintf(stderr,"PAGING raise_and_set_focus done.\n");
}

/*---------- pager ----------*/

typedef struct {
  size_t len;
  unsigned char d[];
} Snapshot;

static void send_key(KeySym sym) {
  int r;
  XTestFakeKeyEvent(disp, keycode(sym),1, 10);
  XTestFakeKeyEvent(disp, keycode(sym),0, 10);
  r= XSync(disp, False);  eassert(r);
}

static void send_pgup_many(void) {
  int i;
  fprintf(stderr,"PAGING   PageUp\n");
  for (i=0; i<25; i++)
    send_key(XK_Prior);
}
static void send_pgdown(void) {
  fprintf(stderr,"PAGING   PageDown\n");
  send_key(XK_Next);
}

static double timestamp(void) {
  struct timeval tv;
  int r;
  
  r= gettimeofday(&tv,0);  eassert(!r);
  double t= (tv.tv_sec - tv_startup.tv_sec) +
            (tv.tv_usec - tv_startup.tv_usec) * 1e-6;
  fprintf(stderr,"%f\n",t);
  return t;
}

static double worst_snapshot= 0.050;
static double worst_stabilise= 0.100;
static size_t snapshot_alloc= 1024;

static void snapshot(Snapshot **output) {
  char *cmd;
  int r;
  
  free(*output);  *output=0;
  double a= timestamp();

  fprintf(stderr,"PAGING   snapshot\n");

  r= asprintf(&cmd, "xwd -silent -id 0x%lx", (unsigned long)id); eassert(r>=0);
  FILE *f= popen(cmd,"r"); eassert(f);

  int need_alloc=1;
  size_t used=0;
  for (;;) {
    size_t allow= snapshot_alloc - used;
    if (!allow) {
      snapshot_alloc <<= 1;
      need_alloc= 1;
      fprintf(stderr,"PAGING   snapshot   grow %ld\n", (long)snapshot_alloc);
      continue;
    }
    if (need_alloc) {
      *output= realloc(*output, sizeof(**output) + snapshot_alloc);
      eassert(*output);
      need_alloc= 0;
    }
    size_t got= fread((*output)->d, 1, allow, f);
    if (got==0) break;
    used += got;
  }
  check_pclose(f,cmd);

  (*output)->len= used;
  snapshot_alloc= used+1;
  *output= realloc(*output, sizeof(**output) + snapshot_alloc);
  eassert(*output);
  
  double b= timestamp();
  double it_took= b-a;
  if (it_took > worst_snapshot)
    worst_snapshot= it_took;

  fprintf(stderr,"PAGING   snapshot took=%f len=%ld\n",
	  it_took, (long)used);
}

static void snapshot_idle(void) {
  int r;
  r= usleep(worst_snapshot * 1e6 * 0.5);
  /* spend no more than 2/3 of our time constantly snapshotting */
  eassert(!r);
}

static int identical(const Snapshot *a, const Snapshot *b) {
  return !(memcmp(a,    b,    sizeof(*a)) ||
	   memcmp(a->d, b->d, a->len));
}

static void wait_for_stability(Snapshot **output,
			       const Snapshot *previously,
			       void (*with_keypress)(void)) {
  Snapshot *last=0;
  /* waits longer if we're going to return an image identical to previously
   * if previously==0, all images are considered identical to it */

  double when_started= timestamp();
  double last_change= when_started;
  double it_took= -1;

  fprintf(stderr,"PAGING  wait_for_stability"
	  "  worst_snapshot=%f worst_stabilise=%f  previously=%p\n",
	  worst_snapshot, worst_stabilise, previously);

  for (;;) {
    double at_snapshot= timestamp();
    snapshot(output);
    double right_now= timestamp();

    if (!last) {
      fprintf(stderr,"PAGING  wait_for_stability first...\n");
      last_change= when_started= right_now;
      last=*output; *output=0;
    } else if (!identical(*output,last)) {
      fprintf(stderr,"PAGING  wait_for_stability changed...\n");
      last_change= right_now;
      free(last); last=*output; *output=0;
      it_took= -1;
    } else {
      if (it_took<0)
	it_took= at_snapshot - when_started;

      double threshold= worst_stabilise + worst_snapshot;
      if (!previously || identical(*output,previously))
	threshold *= 2;  threshold += 1.5;

      double stable_for= right_now - last_change;
      fprintf(stderr,"PAGING  wait_for_stability"
	      "  worst_snapshot=%f worst_stabilise=%f"
	      "  it_took=%f stable_for=%f threshold=%f...\n",
	      worst_snapshot, worst_stabilise,
	      it_took, stable_for, threshold);
      
      if (stable_for > threshold)
	break;
    }
    if (with_keypress)
      with_keypress();
    snapshot_idle();
  }

  if (!with_keypress && it_took > worst_stabilise)
    worst_stabilise= it_took;

  free(last);
  fprintf(stderr,"PAGING  wait_for_stability done.\n");
}

static void read_pages(void) {
  Snapshot *current=0, *last=0;

  raise_and_set_focus();

  /* page to the top - keep pressing page up until the image stops changing */
  wait_for_stability(&last,0, send_pgup_many);

  /* now to actually page down */
  for (;;) {
    fprintf(stderr,"PAGING page %d\n",npages);
    //eassert(npages < MAX_PAGES);
    //page_images[npages++]=
    npages++;

    wait_for_stability(&current,last, 0);
    if (npages &&  /* first pagedown doesn't do much */
	identical(current,last)) {
      free(current);
      break;
    }

    free(last); last=current; current=0;

    send_pgdown();
  }
  fprintf(stderr,"PAGING all done.\n");
}    

int main(int argc, char **argv) {
  screenshot_startup();

  id= strtoul(*++argv,0,0);

  read_pages();
  return 0;
}

